# xDevs.com Calibration app for Keithley 2400, v0.5, Dec.13.2023
# https://xdevs.com/fix/kei2400
# https://xdevs.com/fix/kei2400pa
# https://xdevs.com/fix/hp3458a
import vxi11
import sys
import time
import signal
import numbers
from colorama import init
from math import log10

init(convert=True)                      # Fix ESCAPE-commands support issues with windows

# Add better logging
# Put reading error into function
# Parametrize NPLC, delays, sample time
# Add more verbose data
# Add median/averaging math for readings?
# Show better stats / GUI?
# Eventually refactor calibration function into modules and functions for easier code structure

###################### Configuration settings for script #############################################################

dmm_temp = 23.0
logfilen = 'callog_ver3_2400_dec13_2023.txt'    # Filename for datalog and command reports
samples_num = 10                                # Number of samples to take a average from 3458A
dmm_nplc = 20                                   # NPLC for DMM
run_acal = False                                # If "True" then execute ACAL ALL on 3458A before calibration
gpib_addr_k2400 = 24                            # GPIB address of HP 3458A
gpib_addr_dmm = 3                               # GPIB address of DUT K2400
gpib_vxi_ip = "192.168.1.12"                    # TCP/IP address of Agilent E5810A GPIB-LAN gateway
skip_pre_calibration = False                    # Do not perform calibration before adjustment if True
skip_post_calibration = False                   # Do not perform calibration after adjustment if True

#####################################################################################################################

sys.stdout.write("\033[0;36m Keithley 2400 calibration tool \r\n  Using VXI LAN-GPIB adapter with next instruments config: \r\n * GPIB %d : Keithley 2400 SMU \r\n * GPIB %d : HP 3458A 8.5-digit DMM \r\n ! VXI bridge IP : %s \r\n Do not swap terminals during calibration, high voltage may be present !\033[0;39m\r\n\r\n" % (gpib_addr_k2400, gpib_addr_dmm, gpib_vxi_ip))

dmm_val = 1e-9 # Temp value

class Timeout():
  """Timeout class using ALARM signal"""
  class Timeout(Exception): pass

  def __init__(self, sec):
    self.sec = sec

  def __enter__(self):
    time.sleep(0.5)
    #signal.signal(signal.SIGALRM, self.raise_timeout)
    #signal.alarm(self.sec)

  def __exit__(self, *args):
    time.sleep(0.5)
    #signal.alarm(0) # disable alarm

  def raise_timeout(self, *args):
    raise Timeout.Timeout()

class dmm():
    global dmm_val 
    data = ""
    temp = 26.5
    
    def __init__(self,gpib,name):
        self.gpib = gpib
#        self.inst = Gpib.Gpib(0,self.gpib, timeout=60) # 3458A GPIB Address = self.gpib
        self.inst = vxi11.Instrument(gpib_vxi_ip, "gpib0,%d" % gpib_addr_dmm) # 3458A vxi11 IP address and GPIB address
        self.name = name
        self.init_inst()

    def init_inst(self):
        # Setup HP 3458A
        self.inst.write ("PRESET NORM")
        self.inst.write ("OFORMAT ASCII")
        self.inst.write ("FUNC DCV,AUTO")
        self.inst.write ("TARM HOLD")
        self.inst.write ("TRIG LINE")
        self.inst.write ("NPLC %.3f" % dmm_nplc)
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("MEM OFF")
        self.inst.write ("END ALWAYS")
        self.inst.write ("NDIG 9")
        
    def set_dcv_range(self, cmd):
        self.inst.write ("NPLC %.3f" % dmm_nplc)
        self.inst.write ("DCV %.6e" % cmd)

    def set_dci_range(self, cmd):
        self.inst.write ("NPLC %.3f" % dmm_nplc)
        self.inst.write ("DCI %.6e" % cmd)
        
    def switch_dci(self):
        self.inst.write ("PRESET NORM")
        self.inst.write ("NPLC %.3f" % dmm_nplc)
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("OFORMAT ASCII")
        self.inst.write ("FUNC DCI,AUTO")
        self.inst.write ("NDIG 9")
    
    def write(self,cmd):
        self.inst.write(cmd)
        
    def read(self):
        return self.inst.read()
        
    def math(self,cmd):
        # Setup HP 3458A
        self.inst.write ("MATH %s" % cmd)
    
    def read_data(self,cmd):
        data_float = 0.0
        data_str = ""
        self.inst.write(cmd)
        
        try:
            with Timeout(20):
                data_str = self.inst.read()
        except Timeout.Timeout:
            print ("Timeout exception from dmm %s on read_data() inst.read()\n" % self.name)
            return (0,float(0))
        #print ("Reading from dmm %s = %s" % (self.name,data_str))
        try:
            data_float = float(data_str)
        except ValueError:
            print("Exception thrown by dmm %s on read_data() - ValueError = %s\n" % (self.name,data_str))
            return (0,float(0)) # Exception on float conversion, 0 = error
        return (1,data_float) # Good read, 1 = converted to float w/o exception
        
    def get_temp(self):
        global dmm_temp
        print("Reading DMM temp\r")
        self.inst.write("TARM SGL,1")
        self.temp_status_flag,temp = self.read_data("TEMP?")
        if (self.temp_status_flag):
            self.temp = temp
        dmm_temp = self.temp
        return self.temp
        
    def get_temp_status(self):
        return self.temp_status_flag

    def get_data(self):
        global dmm_val
        self.status_flag,data = self.read_data("TARM SGL,1")
        if (self.status_flag):
            self.data = data
            dmm_val = float(data)
        return self.data
        
    def exec_acal(self):
        sys.stdout.write ("\033[1;35mACAL ALL procedure start, please wait 15 minutes ")
        self.inst.write("ACAL ALL")
        for cnt in range(0 ,290):        # wait 870 seconds, print dot every 3s
            sys.stdout.write ("\033[0;40m*")
            time.sleep(3)
            sys.stdout.flush()
        print("\rACAL procedure done...\033[1;39m")
    
    def exec_idn(self):
        self.inst.write ("END ALWAYS")
        self.inst.write ("ID?")
        dat = self.inst.read()
        tstr = dat.split()
        if (tstr[0] == "HP3458A"):
            sys.stdout.write ("\r\033[1;32m%s detected...\033[1;39m" % tstr[0])
        else:
            sys.stdout.write ("\r\033[1;31mNo DMM present, exiting!\033[1;39m")
            quit()

    def get_data_status(self):
        return self.status_flag

def perftest():
    dmm.init_inst()
    dmm.get_temp()
    print ("\r\n\033[0;33mHP3458A TEMP = %2.1f C\033[0;33m" % dmm_temp)
    
    for cnt in range (0, len(test_dcv_set)):
        kei2400.write (":SOUR:FUNC VOLT")
        dmm.set_dcv_range(test_dcv_rng[cnt])
        kei2400.write (":SENS:CURR:PROT 0.01")
        kei2400.write (":SENS:CURR:RANG 0.01")
        kei2400.write (":OUTP:STAT ON")
        kei2400.write (":SOUR:VOLT:RANGE %3.1e" % test_dcv_rng[cnt])
        kei2400.write (":SOUR:VOLT %3.1e" % test_dcv_set[cnt])
        time.sleep(1)
    
        val = 0
        if (abs(test_dcv_set[cnt]) > 100):
            # Perform additional dummy 60 measurement steps to allow for dividers in DUT and DMM warm-up for >100 V points
            for samples in range (0, 60):
                dmm.get_data()
                print ("Warmup DCV %d : %.9E VDC\r" % (samples, dmm_val))
                
        for samples in range (0, samples_num):
            dmm.get_data()
            print ("Test DCV %d : %.9E VDC\r" % (samples, dmm_val))
            val += dmm_val
        val = val / samples_num
        ppm = (((val+1e-16) / (test_dcv_set[cnt] + 1E-16)) - 1) * 1E6
        
        sys.stdout.write(" Verification step  %02d (%.9E) : %.9E VDC [deviation %6.2f ppm]\r\n" % (cnt, test_dcv_set[cnt], val, ppm))
        b.write("DCV Verification step %02d (%.9E), measured : %.9E VDC, deviation %.3f ppm\r" % (cnt, test_dcv_set[cnt], val, ppm))
        cnt = cnt + 1
    
    print ("\033[1;31m\r\nConnect DMM CURRENT input to Model 2400 INPUT/OUTPUT jacks.\033[1;39m")
    input("\033[1;33mPress Enter to continue with calibration...\033[1;39m")
    
    sys.stdout.write("\r\n\r\nPerformance DCI test\r\n")
    dmm.init_inst()
    dmm.get_temp()
    dmm.switch_dci()
    sys.stdout.write("3458A TEMP? = %02.1f \r\n" % (dmm_temp))
    
    for cnt in range (0, len(test_dci_set)):
        kei2400.write (":SOUR:FUNC CURR")
        dmm.set_dci_range(test_dci_rng[cnt])
        kei2400.write (":SENS:VOLT:PROT 10")
        kei2400.write (":SENS:VOLT:RANG 10")
        kei2400.write (":OUTP:STAT ON")
        kei2400.write (":SOUR:CURR:RANGE %3.1e" % test_dci_rng[cnt])
        kei2400.write (":SOUR:CURR %3.1e" % test_dci_set[cnt])
        time.sleep(3)
    
        if (abs(test_dci_set[cnt]) > 200e-3):
            # Perform additional dummy 60 measurement steps to allow for shunts in DUT and DMM warm-up for 1 A points
            for samples in range (0, 60):
                dmm.get_data()
                print ("Warmup DCI %d : %.9E VDC\r" % (samples, dmm_val))

        val = 0
        for samples in range (0, samples_num):
            dmm.get_data()
            print ("Test DCI %d : %.9E ADC" % (samples, dmm_val))
            val += dmm_val
        val = val / samples_num
        ppm = (((val+1e-16) / (test_dci_set[cnt] + 1E-16)) - 1) * 1E6
        sys.stdout.write(" Verification step %sA : %.9E ADC [deviation %6.2f ppm]\r\n" % (test_dci_str[cnt], val, ppm))
        b.write("DCI Verification step %sA , measured : %.9E ADC, deviation %.3f ppm\r" % (test_dci_str[cnt], val, ppm))
        cnt = cnt + 1
    
    kei2400.write (":OUTP:STAT OFF")
    sys.stdout.write("\033[0;31;43mConnect DMM VOLTAGE input to 2400 INPUT/OUTPUT jacks.\033[0;39;49m\r\n\r\n")
    print (":SYST:BEEP:IMM 1000, 1")
    print (":SYST:BEEP:IMM 2000, 2")
    print (":SYST:BEEP:IMM 5000, 5")
    input("\r\nPress Enter to continue to next step...")
    

with open(logfilen,'a') as b:
#    kei2400 = Gpib.Gpib(0,24)               # Keithley Model 2400 GPIB address for use with linux-gpib
    b.write ("\r\n-i- Script settings : %d samples, %d DMM NPLC, %d ACAL run, skip_precal %d, skip_postcal %d\r\n" % (samples_num, dmm_nplc, run_acal, skip_pre_calibration, skip_post_calibration))
    timstr = time.strftime("%d/%m/%Y-%H:%M:%S;")
    b.write ("-i- Script started %s" % timstr)
    
    kei2400 = vxi11.Instrument(gpib_vxi_ip, "gpib0,%d" % gpib_addr_k2400) # Keithley 2400 vxi11 IP address and GPIB address
    kei2400.clear()

    # Adjustment points per Keithley service manual
    cal_dcv_set = [-0.2,   0, 0.2,   0, -2, 0, 2, 0, -20,  0, 20,  0, -200,   0, 200,   0]
    cal_dcv_rng = [ 0.2, 0.2, 0.2, 0.2,  2, 2, 2, 2,  20, 20, 20, 20,  200, 200, 200, 200]
    cal_dcv_str = ["-0.2 "," -Zero"," 0.2 "," +Zero","   -2 "," -Zero","    2 "," -Zero","  -20 "," +Zero","   20 "," -Zero"," -200 ","-Zero","  200 "," +Zero"]
    cal_dci_set = [-1E-6,    0, 1E-6,    0, -10E-6,     0, 10E-6,     0, -100E-6,      0, 100E-6,      0, -1E-3,    0, 1E-3,    0,  -10E-3,     0, 10E-3,     0, -100E-3,      0, 100E-3,      0, -1, 0, 1, 0]
    cal_dci_rng = [ 1E-6, 1E-6, 1E-6, 1E-6,  10E-6, 10E-6, 10E-6, 10E-6,  100E-6, 100E-6, 100E-6, 100E-6,  1E-3, 1E-3, 1E-3, 1E-3,   10E-3, 10E-3, 10E-3, 10E-3,  100E-3, 100E-3, 100E-3, 100E-3,  1, 1, 1, 1]
    cal_dci_str = ["-1 u","-Zero", " 1 u", "+Zero", " -10 u", "-Zero", " 10 u", "+Zero", "-100 u", "-Zero", " 100 u", "+Zero", "-1 m", "-Zero", " 1 m", "+Zero", " -10 m", "-Zero", " 10 m", "+Zero", "-100 m", "-Zero", " 100 m", "+Zero", "-1 ", "-Zero ", "-1 ", "+Zero "]
    # Calibration performance test points
    test_dcv_set = [0, -0.002, 0.002, -0.02, 0.02, -0.2, 0.2,   0, -1, 1, -2, 2, 0, -10, 10, -15, 15, -20, 20,  0, -100, 100,-200, 200,  0]
    test_dcv_rng = [0.1,  0.1,   0.1,   0.1,  0.1,  0.2, 0.2, 0.2,  1, 1,  2, 2, 2,  10, 10,  20, 20,  20, 20, 10,  100, 100, 200, 200,100]
    test_dci_set = [   0, -1E-9, 1E-9, -10E-9, 10E-9, -100E-9, 100E-9, -1E-6, 1E-6, -10E-6, 10E-6, -100E-6, 100E-6, -1E-3, 1E-3, -4E-3,  4E-3, -10E-3, 10E-3, -20E-3,  20E-3,  -50E-3,  50E-3, -100E-3, 100e-3, -1, 1, 0]
    test_dci_rng = [1E-6,  1E-6, 1E-6,   1E-6,  1E-6,    1E-6,   1E-6,  1E-6, 1E-6,  10E-6, 10E-6,  100E-6, 100E-6,  1E-3, 1E-3, 10E-3, 10E-3,  10E-3, 10E-3, 100E-3, 100E-3,  100E-3, 100E-3,  100e-3, 100e-3,  1, 1, 1]
    test_dci_str = ["Zero  ","  -1 n", "   1 n", " -10 n", "  10 n", "-100 n", " 100 n", "  -1 u", "   1 u", " -10 u", "  10 u", "-100 u", " 100 u", "  -1 m", "   1 m", "  -4 m", "   4 m", " -10 m", "  10 m", " -20 ", "  20 m", "-100 m", " 100 m", "-200 m", " 200 m", "-1 ", "1 ", "Zero  "]
    kei2400_idn = "KEITHLEY INSTRUMENTS INC."
    kei2400_mdl = "MODEL 2400"
    orig_val = []
    kei_error = 0
    plc = 60 #60 Hz
    cnt = 0
    samples = 0
    val = 3.423
    meas = 0.0
    temp = 36.6
    
    for cnt in range(30):
        orig_val.append(0)
            
    kei2400.write ("*IDN?")
    idn_str = kei2400.read()
    idmfg = idn_str.split(",")
    if (kei2400_idn == idmfg[0]):
        if (kei2400_mdl == idmfg[1]):
            sys.stdout.write("\033[0;32mKeithley Model 2400 - detected, S/N %s, Version: %s\033[0;39m\r\n" % (idmfg[2], idmfg[3]))
            b.write("\r\nKeithley Model 2400 - detected, S/N %s, Version: %s\r" % (idmfg[2], idmfg[3]))
        else:
            print ("\033[0;31mIncorrect Model! Check GPIB address. Calibration cancelled.\033[0;39m")    
            quit()
    else:
        sys.stdout.write ("\033[0;31mNo Keithley instrument detected. Check GPIB address. Calibration cancelled.\033[0;39m")
        quit()
    
    # Keithley Model 2400 detected OK, proceed with reading CAL data
    kei2400.write (":CAL:PROT:DATE?")
    last_cal_date = kei2400.read()
    b.write("\r\nKeithley Model 2400 - last date calibrated %s \r" % last_cal_date)
    kei2400.write (":CAL:PROT:NDUE?")
    last_due_date = kei2400.read()
    b.write("\r\nKeithley Model 2400 - last due calibrated %s \r" % last_due_date) 
 
    dmm = dmm(3,"3458A") # GPIB 3
    dmm.exec_idn()
    dmm.get_temp()
    dmm.math('OFF')
    
    print ("\033[0;33m\r\nHP3458A TEMP = %2.1f C\033[0;33m" % dmm_temp)
    b.write ("\r\nHP 3458A detected\r TEMP? = %2.1f C\r" % dmm_temp)

    dmm.write ("CAL? 72")
    dmm_cal72 = dmm.read()
    b.write("\r\nHP3458 - CAL72 value %s" % dmm_cal72)
    dmm.write ("CAL? 1,1")
    dmm_cal11 = dmm.read()
    b.write("\r\nHP3458 - CAL1,1 value %s" % dmm_cal11)
    dmm.write ("CAL? 2,1")
    dmm_cal21 = dmm.read()
    b.write("\r\nHP3458 - CAL2,1 value %s" % dmm_cal21)
 
    kei2400.write ("*RST")
    kei2400.write ("*CLS")
    kei2400.write ("*ESE 1;*SRE 32")
    kei2400.write (":SENS:FUNC:CONC OFF")
    kei2400.write (":SENS:FUNC 'VOLT:DC'")
    kei2400.write (":SOUR:FUNC VOLT")
    kei2400.write (":OUTP:STAT OFF")
    
    print ("Reading Pre-cal calibration data from Model 2400")
    b.write("Pre-cal K2400 Calibration data\r")
    
    for cnt in range (0, 4):
        kei2400.write (":SOUR:VOLT:RANGE %3.1f" % cal_dcv_rng[cnt*4])
        kei2400.write (":CAL:PROT:SENS:DATA?")
        kdata = kei2400.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1f DCV Range, SENS:DATA? = %s\r" % (cal_dcv_rng[cnt*4], kdata))
        kei2400.write (":CAL:PROT:SOUR:DATA?")
        kdata = kei2400.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1f DCV Range, SOUR:DATA? = %s\r" % (cal_dcv_rng[cnt*4], kdata))
        cnt = cnt + 1
    
    kei2400.write (":CAL:PROT:LOCK")
    kei2400.write (":SENS:FUNC 'CURR:DC'")
    kei2400.write (":SOUR:FUNC CURR")
    
    for cnt in range (0, 7):
        kei2400.write (":SOUR:CURR:RANGE %3.1e" % cal_dci_rng[cnt*4])
        kei2400.write (":CAL:PROT:SENS:DATA?")
        kdata = kei2400.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1e DCI Range, SENS:DATA? = %s\r" % (cal_dci_rng[cnt*4], kdata))
        kei2400.write (":CAL:PROT:SOUR:DATA?")
        kdata = kei2400.read()
        b.write("%3.1e DCI Range, SOUR:DATA? = %s\r" % (cal_dci_rng[cnt*4], kdata))
        sys.stdout.write ("%s" % kdata)
        cnt = cnt + 1

    # K2400 data read OK. Detect HP 3458A DMM on GPIB 3

    kei2400.write (":SYST:BEEP:IMM 800, 5")
    kei2400.write (":SYST:BEEP:IMM 1400, 3")
    kei2400.write (":SYST:BEEP:IMM 2200, 2")
    kei2400.write (":SYST:BEEP:IMM 3000, 1")
    print ("\033[1;31m\r\nConnect DMM volts input to Model 2400 INPUT/OUTPUT jacks.\033[1;39m")
    input("\033[1;33mPress Enter to continue with calibration...\033[1;39m")
    
    if (run_acal == True):
        dmm.exec_acal()         # Execute ACAL on HP 3458A for best accuracy. Takes 15 minutes to complete.
    
    kei2400.write ("*RST")
    kei2400.write (":SOUR:FUNC VOLT")
    kei2400.write (":SENS:CURR:PROT 0.01")
    kei2400.write (":SENS:CURR:RANG 0.01")
    kei2400.write (":SOUR:VOLT:PROT:LEV MAX")
    kei2400.write (":SYST:RSEN OFF")
    
    if (skip_pre_calibration != True):
        sys.stdout.write("\r\nPerformance DCV test\r\n")
        perftest()                                                  # Run pre-adjustment calibration 
    
    adj_ready = input("Start adjustment procedure? Enter YES to continue, or anything else to skip... :")
    if (adj_ready == 'YES'):
        kei2400.write (":CAL:PROT:CODE 'KI002400'")
        dmm.init_inst()
        dmm.get_temp()
        print ("\r\n\033[0;33mHP3458A TEMP = %2.1f C\033[0;33m" % dmm_temp)
        kei2400.write ("*RST")
        kei2400.write ("*CLS")
        kei2400.write ("*ESE 1;*SRE 32")
        kei2400.write (":CAL:PROT:CODE 'KI002400'")
        kei2400.write (":OUTP:STAT ON")
        sys.stdout.write("Starting DCV adjustment...\r\n")
    
        for cnt in range (0, len(cal_dcv_set)):
            sys.stdout.write("\r Calibration step %d " % cnt)
            dmm.set_dcv_range(cal_dcv_rng[cnt])
            kei2400.write (":SOUR:VOLT:RANGE %3.1f" % cal_dcv_rng[cnt])
            kei2400.write (":SOUR:VOLT %3.1f" % cal_dcv_set[cnt])
            b.write (":SOUR:VOLT:RANGE %3.1f\r\n" % cal_dcv_rng[cnt])
            b.write (":SOUR:VOLT %3.1f\r\n" % cal_dcv_set[cnt])
            
            if (abs(cal_dcv_set[cnt]) > 100):
                # Perform additional dummy 60 measurement steps to allow for dividers in DUT and DMM warm-up for >100 V points
                for samples in range (0, 60):
                    dmm.get_data()
                    print ("Warmup DCV cal %d : %.7f VDC\r" % (samples, dmm_val))
            
            val = 0 
            time.sleep(1)
            for samples in range (0, samples_num):
                dmm.get_data()
                print ("\rMeasured DCV %d : %.7f VDC" % (samples, dmm_val))
                val += dmm_val
            val = val / samples_num
            
            orig_val[cnt] = val
            sys.stdout.write("\rDCV Calibration step %sV : %.9E VDC" % (cal_dcv_str[cnt], val))
            
            kei2400.write (":CAL:PROT:SOUR %3.6f;*OPC" % val)
            time.sleep(3)
            b.write (":CAL:PROT:SOUR %3.5f;*OPC\r\n" % val)
            kei2400.write (":SYST:ERR?")
            kei_error = kei2400.read().split(",")
            if (int(kei_error[0]) != 0):
                print ("\r\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0]), kei_error, "\r\n")
                quit()
            #kei2400.write ("*STB?")
            #kei2400.write ("*ESR?")
            
            kei2400.write (":CAL:PROT:SENS %3.6e;*OPC" % val)
            time.sleep(3)
            b.write (":CAL:PROT:SENS %3.5f;*OPC\r\n" % val)
            kei2400.write (":SYST:ERR?")
            kei_error = kei2400.read().split(",")
            if (int(kei_error[0]) != 0):
                print ("\r\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0]))
                quit()
            #kei2400.write ("*STB?")
            #kei2400.write ("*ESR?")
            
            if (cnt % 2 == 0):
                ppms = ((val / (cal_dcv_set[cnt] + 1E-16)) - 1) * 1E6
                sys.stdout.write("\r\n\033[1;35m Delta: %.3f ppm\033[1;39m\r" % ppms)
                b.write("DCV Calibration step %sV , measured : %.9E VDC, deviation %.3f ppm\r" % (cal_dcv_str[cnt], val, ppms))
            else:
                sys.stdout.write("\r\n")
            cnt = cnt + 1
    
        sys.stdout.write("\r\n")
        sys.stdout.write("\033[0;31;43mConnect DMM CURRENT input to 2400 INPUT/OUTPUT jacks.\033[0;39;49m\r\n")
        
        kei2400.write (":SYST:BEEP:IMM 1000, 5")
        input("Press Enter to continue...")
        
        sys.stdout.write("\rStarting DCI adjustment...\r\n\r\n")
        
        dmm.switch_dci()
    
        kei2400.write (":SOUR:FUNC CURR")
        kei2400.write (":SENS:VOLT:PROT 20")
        kei2400.write (":SENS:VOLT:RANG 20")
        kei2400.write (":OUTP:STAT ON")
        
        for cnt in range (0, len(cal_dci_set)):
            kei2400.write (":SOUR:CURR:RANGE %3.6e" % cal_dci_rng[cnt])
            kei2400.write (":SOUR:CURR %3.6e" % cal_dci_set[cnt])
            print ("Setting DMM DCI range %3.3e A" % cal_dci_rng[cnt])
            dmm.set_dci_range(cal_dci_rng[cnt])
            b.write (":SOUR:CURR:RANGE %3.6e\r\n" % cal_dci_rng[cnt])
            b.write (":SOUR:CURR %3.6e\r\n" % cal_dci_set[cnt])
            time.sleep(3) # Soak 3 seconds
            val = 0 
            
            if (abs(cal_dci_set[cnt]) > 200e-3):
                # Perform additional dummy 60 measurement steps to allow for shunts in DUT and DMM warm-up for 1 A points
                for samples in range (0, 60):
                    dmm.get_data()
                    print ("Warmup cal DCI %d : %.9E VDC\r" % (samples, dmm_val))
            
            for samples in range (0, samples_num):
                dmm.get_data()
                print ("Measured DCI %d : %.7f ADC" % (samples, dmm_val))
                val += dmm_val
            val = val / samples_num
    
            sys.stdout.write("\rDCI Calibration step %sA : %.6E ADC" % (cal_dci_str[cnt], val))
            
            kei2400.write (":CAL:PROT:SOUR %3.6e;*OPC" % val)
            time.sleep(1)
            b.write (":CAL:PROT:SOUR %3.6e;*OPC\r\n" % val)
            kei2400.write (":SYST:ERR?")
            kei_error = kei2400.read().split(",")
            if (int(kei_error[0]) != 0):
                print ("\r\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0]))
                quit()
            #kei2400.write ("*STB?")
            #kei2400.write ("*ESR?")
            
            kei2400.write (":CAL:PROT:SENS %3.6e;*OPC" % val)
            b.write (":CAL:PROT:SENS %3.6e;*OPC\r\n" % val)
            time.sleep(1)
            kei2400.write (":SYST:ERR?")
            kei_error = kei2400.read().split(",")
            if (int(kei_error[0]) != 0):
                print ("\033[1;31mKeithley 2400 error : %d\033[1;39m" % int(kei_error[0]))
                quit()
            #kei2400.write ("*STB?")
            #kei2400.write ("*ESR?")
            
            if (cnt % 2 == 0):
                ppms = ((val / (cal_dci_set[cnt] + 1E-16)) - 1) * 1E6
                sys.stdout.write("\033[1;36m Delta: %.3f ppm\033[1;39m\r\n" % ppms)
                b.write("DCI Calibration step %sA , measured : %.6E ADC, deviation %.1f ppm\r" % (cal_dci_str[cnt], val, ppms))
            else:
                sys.stdout.write("\r\n")
            cnt = cnt + 1
    
        kei2400.write (":CAL:PROT:DATE 2023, 12, 12")
        kei2400.write (":CAL:PROT:NDUE 2024, 12, 12")
        kei2400.write (":CAL:PROT:SAVE")
        kei2400.write (":CAL:PROT:LOCK")
        kei2400.write (":OUTP:STAT OFF")
        b.write (":CAL:PROT:DATE 2023, 12, 12\r\n")
        b.write (":CAL:PROT:NDUE 2024, 12, 12\r\n")
        b.write (":CAL:PROT:SAVE\r\n")
        b.write (":CAL:PROT:LOCK\r\n")
        b.write (":OUTP:STAT OFF\r\n")
    
        print("\r\033[0;30;42mAdjustment complete! www.xDevs.com \033[0;39;49m")
        kei2400.write (":SYST:BEEP:IMM 100, 1")
        kei2400.write (":SYST:BEEP:IMM 300, 1")
        kei2400.write (":SYST:BEEP:IMM 500, 1")
        kei2400.write (":SYST:BEEP:IMM 900, 1")
        kei2400.write (":SYST:BEEP:IMM 1200, 1")
        kei2400.write (":SYST:BEEP:IMM 1500, 1")
        kei2400.write (":SYST:BEEP:IMM 2000, 1")
    
        if (skip_post_calibration != True):
            print ("\033[1;31m\r\nConnect DMM Volts input to Model 2400 INPUT/OUTPUT jacks.\033[1;39m")
            input("\033[1;33mPress Enter to continue with calibration...\033[1;39m")
            sys.stdout.write("\r\n\r\nPost-cal Performance calibration test\r\n")
            perftest()                                      # Run post-adjustment performance test
    else:
        sys.stdout.write("\r\n\r\nAdjustment and post-adjustment calibration is skipped...\r\n")
    
    # Post-cal constants 
    print ("Reading Post-cal calibration data from Model 2400")
    b.write("Post-cal K2400 Calibration data\r")
    
    for cnt in range (0, 4):
        kei2400.write (":SOUR:VOLT:RANGE %3.1f" % cal_dcv_rng[cnt*4])
        kei2400.write (":CAL:PROT:SENS:DATA?")
        kdata = kei2400.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1f DCV Range, SENS:DATA? = %s\r" % (cal_dcv_rng[cnt*4], kdata))
        kei2400.write (":CAL:PROT:SOUR:DATA?")
        kdata = kei2400.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1f DCV Range, SOUR:DATA? = %s\r" % (cal_dcv_rng[cnt*4], kdata))
        cnt = cnt + 1
    
    kei2400.write (":SENS:FUNC 'CURR:DC'")
    kei2400.write (":SOUR:FUNC CURR")
    
    for cnt in range (0, 7):
        kei2400.write (":SOUR:CURR:RANGE %3.1e" % cal_dci_rng[cnt*4])
        kei2400.write (":CAL:PROT:SENS:DATA?")
        kdata = kei2400.read()
        sys.stdout.write ("%s" % kdata)
        b.write("%3.1e DCI Range, SENS:DATA? = %s\r" % (cal_dci_rng[cnt*4], kdata))
        kei2400.write (":CAL:PROT:SOUR:DATA?")
        kdata = kei2400.read()
        b.write("%3.1e DCI Range, SOUR:DATA? = %s\r" % (cal_dci_rng[cnt*4], kdata))
        sys.stdout.write ("%s" % kdata)
        cnt = cnt + 1
        
    kei2400.write (":CAL:PROT:LOCK")
    b.write ("\r\nCalibration complete!\r\n")
    timstr = time.strftime("%d/%m/%Y-%H:%M:%S;")
    b.write ("-i- Script sucessfully finished %s" % timstr)
    b.close()
    
    print (":SYST:BEEP:IMM 1000, 1")
    print (":SYST:BEEP:IMM 4000, 2")
    print (":SYST:BEEP:IMM 8000, 2")
    print (":SYST:BEEP:IMM 2000, 2")
    print (":SYST:BEEP:IMM 5000, 2")
    print (":SYST:BEEP:IMM 8000, 2")
    print (":SYST:BEEP:IMM 2000, 5")
    print (":SYST:BEEP:IMM 7000, 5")
    print (":SYST:BEEP:IMM 9000, 5")

    print("\r\nAll done!")